{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Linear classification with rate constraints\n\nIn this example we consider a linear classification problem on synthetically\nmixture of Gaussians data. We constraint the model to predict at least 70% of\nthe training points as blue.\n\nNote that this is a non-differentiable constraint, and thus the typical\nLagrangian approach is not applicable as it requires to compute the derivatives\nof the constraints $g$ and $h$.\n\nA commonly used approach to deal with this difficulty is to retain the\nLagrangian formulation, but replace the constraints with differentiable\napproximations or surrogates. However, changing the constraint functions can\nresult in an over- or under-constrained version of the problem.\n\n:cite:t:`cotter2019JMLR` propose a *proxy-Lagrangian formulation*, in which the\nnon-differentiable constraints are relaxed *only when necessary*. In other\nwords, the non differentiable constraint functions are used to compute the\nLagrangian and constraint violations (and thus the multiplier updates), while\nthe surrogates are used to compute the primal gradients.\n\nThis example is based on Fig. 2 of :cite:t:`cotter2019JMLR`. Here we present\nthe naive setting where proxy-constraints are used on a Lagrangian formulation,\nrather than a \"proper\" proxy-Lagrangian formulation. For details on the\nnotion of a proxy-Lagrangian formulation, see $\\S$ 4.2 in\n:cite:p:`cotter2019JMLR`.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import random\n\nimport matplotlib.pyplot as plt\nimport numpy as np\nimport style_utils\nimport torch\nfrom torch.nn.functional import binary_cross_entropy_with_logits as bce_loss\n\nimport cooper\n\ntorch.manual_seed(0)\nnp.random.seed(0)\nrandom.seed(0)\n\n\ndef generate_mog_dataset():\n    \"\"\"\n    Generate a MoG dataset on 2D, with two classes.\n    \"\"\"\n\n    n_per_class = 100\n    dim = 2\n    n_gaussians = 4\n    mus = [(0, 1), (-1, 0), (0, -1), (1, 0)]\n    mus = [torch.tensor(m) for m in mus]\n    var = 0.05\n\n    inputs, labels = [], []\n\n    for id in range(n_gaussians):\n        # Generate input data by mu + x @ sqrt(cov)\n        cov = np.sqrt(var) * torch.eye(dim)\n        mu = mus[id]\n        inputs.append(mu + torch.randn(n_per_class, dim) @ cov)\n\n        # Labels\n        labels.append(torch.tensor(n_per_class * [1.0 if id < 2 else 0.0]))\n\n    return torch.cat(inputs, dim=0), torch.cat(labels, dim=0)\n\n\ndef plot_pane(ax, inputs, x1, x2, achieved_const, titles, colors):\n    const_str = str(np.round(achieved_const, 0)) + \"%\"\n    ax.scatter(*torch.transpose(inputs, 0, 1), color=colors)\n    ax.plot(x1, x2, color=\"gray\", linestyle=\"--\")\n    ax.fill_between(x1, -2, x2, color=blue, alpha=0.1)\n    ax.fill_between(x1, x2, 2, color=red, alpha=0.1)\n\n    ax.set_aspect(\"equal\")\n    ax.set_xlim(-2, 2)\n    ax.set_ylim(-2, 2)\n    ax.set_title(titles[idx] + \" - Pred. Blue Prop.: \" + const_str)\n\n\nclass MixtureSeparation(cooper.ConstrainedMinimizationProblem):\n    \"\"\"\n    Implements CMP for separating the MoG dataset with a linear predictor.\n\n    Args:\n        is_constrained: Flag to apply or not the constraint on the percentage of\n            points predicted as belonging to the blue class\n        use_proxy: Flag to use proxy-constraints. If ``True``, we use a hinge\n            relaxation. Defaults to ``False``.\n        const_level: Minimum proportion of points to be predicted as belonging\n            to the blue class. Ignored when ``is_constrained==False``. Defaults\n            to ``0.7``.\n    \"\"\"\n\n    def __init__(\n        self, is_constrained: bool, use_proxy: bool = False, const_level: float = 0.7\n    ):\n\n        super().__init__(is_constrained=is_constrained)\n\n        self.const_level = const_level\n        self.use_proxy = use_proxy\n\n    def closure(self, model, inputs, targets):\n\n        logits = model(inputs)\n        loss = bce_loss(logits.flatten(), targets)\n\n        if not self.is_constrained:\n            # Unconstrained problem of separating two classes\n            state = cooper.CMPState(\n                loss=loss,\n            )\n        else:\n            # Separating classes s.t. predicting at least const_level as class 0\n\n            # Hinge approximation of the rate\n            probs = torch.sigmoid(logits)\n            hinge = torch.mean(torch.max(torch.zeros_like(probs), 1 - probs))\n\n            # level - proxy_ineq_defect <= 0\n            hinge_defect = self.const_level - hinge\n\n            if not self.use_proxy:\n                ineq_defect = hinge_defect\n                proxy_ineq_defect = None\n            else:\n                # Use non-proxy defects to update the Lagrange multipliers\n\n                # Proportion of elements in class 0 is the non-proxy defect\n                classes = logits >= 0.0\n                prop_0 = torch.sum(classes == 0) / targets.numel()\n                ineq_defect = self.const_level - prop_0\n                proxy_ineq_defect = hinge_defect\n\n            state = cooper.CMPState(\n                loss=loss,\n                ineq_defect=ineq_defect,\n                proxy_ineq_defect=proxy_ineq_defect,\n            )\n\n        return state\n\n\ndef train(problem_name, inputs, targets, num_iters=5000, lr=1e-2, const_level=0.7):\n    \"\"\"\n    Train via SGD\n    \"\"\"\n\n    is_constrained = problem_name.lower() in [\"constrained\", \"proxy\"]\n    use_proxy = problem_name.lower() == \"proxy\"\n\n    model = torch.nn.Linear(2, 1)\n\n    cmp = MixtureSeparation(is_constrained, use_proxy, const_level)\n    formulation = cooper.LagrangianFormulation(cmp)\n\n    primal_optimizer = torch.optim.SGD(model.parameters(), lr=lr, momentum=0.7)\n\n    if is_constrained:\n        dual_optimizer = cooper.optim.partial_optimizer(torch.optim.SGD, lr=lr, momentum=0.7)\n    else:\n        dual_optimizer = None\n\n    constrained_optimizer = cooper.ConstrainedOptimizer(\n        formulation=formulation,\n        primal_optimizer=primal_optimizer,\n        dual_optimizer=dual_optimizer,\n    )\n\n    for i in range(num_iters):\n        constrained_optimizer.zero_grad()\n        if is_constrained:\n            lagrangian = formulation.composite_objective(\n                cmp.closure, model, inputs, targets\n            )\n            formulation.custom_backward(lagrangian)\n        else:\n            # No Lagrangian in the unconstrained case\n            loss = cmp.closure(model, inputs, targets).loss\n            loss.backward()\n\n        constrained_optimizer.step()\n\n    # Number of elements predicted as class 0 in the train set after training\n    logits = model(inputs)\n    pred_classes = logits >= 0.0\n    prop_0 = torch.sum(pred_classes == 0) / targets.numel()\n\n    return model, 100 * prop_0.item()\n\n\n# Plot configs\ntitles = [\"Unconstrained\", \"Constrained\", \"Proxy\"]\nfig, axs = plt.subplots(1, 3, figsize=(18, 6))\n\n# Data and training configs\ninputs, labels = generate_mog_dataset()\nconst_level = 0.7\nlr = 2e-2\nnum_iters = 5000\n\nfor idx, name in enumerate(titles):\n\n    model, achieved_const = train(\n        name, inputs, labels, lr=lr, num_iters=num_iters, const_level=const_level\n    )\n\n    # Compute decision boundary\n    weight, bias = model.weight.data.flatten().numpy(), model.bias.data.numpy()\n    x1 = np.linspace(-2, 2, 100)\n    x2 = (-1 / weight[1]) * (weight[0] * x1 + bias)\n\n    # Color points according to true label\n    red, blue = style_utils.COLOR_DICT[\"red\"], style_utils.COLOR_DICT[\"blue\"]\n    colors = [red if _ == 1 else blue for _ in labels.flatten()]\n    plot_pane(axs[idx], inputs, x1, x2, achieved_const, titles, colors)\n\nfig.suptitle(\"Goal: Predict at least \" + str(const_level * 100) + \"% as blue\")\nplt.show()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.13"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}